"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const bytes_1 = __importDefault(require("bytes"));
const groupBy = (list = [], fn) => {
    const group = new Map();
    for (const x of list) {
        const key = fn(x);
        const value = group.get(key);
        if (value) {
            value.push(x);
        }
        else {
            group.set(key, [x]);
        }
    }
    return group;
};
const splitModuleName = (modulePath = '') => {
    const [p1, p2, ...rest] = modulePath.split(path_1.default.sep);
    return p1.startsWith('@')
        ? [p1 + '/' + p2, rest.join(path_1.default.sep)]
        : [p1, [p2, ...rest].join(path_1.default.sep)];
};
const npmDir = 'node_modules';
const maxFilesInCell = 10;
/**
 * Like `esbuild.analyzeMetafile`, but:
 * - render to markdown table
 * - more compact, accumulate files in same module
 * - strip common ancestor directory
 */
function analyzeMetafile(data) {
    const { outputs } = data;
    const markdown = [];
    for (const [outputName, outputStats] of Object.entries(outputs)) {
        const { entryPoint, inputs, bytes: totalBytes } = outputStats;
        const list = Object.entries(inputs).map(([x, { bytesInOutput }]) => {
            let names;
            let index;
            if (x === entryPoint) {
                names = { parent: '', head: x, file: '' };
            }
            else if ((index = x.indexOf(npmDir)) !== -1) {
                // handle node modules
                const parent = x.slice(0, index);
                const [moduleName, part2] = splitModuleName(x.slice(index + npmDir.length + 1));
                names = { parent, head: moduleName, file: part2 };
            }
            else {
                // handle alias from tsconfig/jsconfig
                const reParent = /^((\.\.[/\\])+)/;
                const [part1, parent, _, rest] = x.split(reParent);
                if (parent) {
                    names = {
                        parent,
                        head: path_1.default.dirname(rest),
                        file: path_1.default.basename(rest),
                    };
                }
                else {
                    names = {
                        parent,
                        head: path_1.default.dirname(part1),
                        file: path_1.default.basename(part1),
                    };
                }
            }
            return { names, bytesInOutput };
        });
        const group = groupBy(list, (x) => x.names.parent);
        const nameGrouped = [...group].map(([_, entries]) => {
            return groupBy(entries, (x) => x.names.head);
            // TODO: show parent dir if needed, alert duplicates
            // return groupBy(entries, (x) => x.names[0] + x.names[1])
        });
        const head = ['Input', 'Files', 'Size', 'Percent'];
        const lines = [
            `| ${head.join(' | ')} |`,
            `| ${head.map(() => '---').join(' | ')} |`,
        ];
        const leadingRow = {
            name: outputName,
            files: '',
            size: totalBytes,
            percent: '100%',
        };
        const rows = [leadingRow];
        for (const map of nameGrouped) {
            for (const [name, group] of map) {
                const files = group
                    .map((x) => ({ name: x.names.file, size: x.bytesInOutput }))
                    .sort((a, b) => b.size - a.size);
                const bytesAcc = group.reduce((acc, x) => acc + x.bytesInOutput, 0);
                if (bytesAcc === 0) {
                    continue;
                }
                const percent = ((bytesAcc / totalBytes) * 100).toFixed(1) + '%';
                const getFiles = () => {
                    if (files.length === 1) {
                        return files[0].name;
                    }
                    const format = (x) => `${x.name} (${(0, bytes_1.default)(x.size)})`;
                    if (files.length > maxFilesInCell) {
                        const rest = files.slice(maxFilesInCell);
                        const restSize = rest.reduce((acc, x) => acc + x.size, 0);
                        return (files.slice(0, maxFilesInCell).map(format).join(', ') +
                            `, and ${rest.length} other ${rest.length > 1 ? 'files' : 'file'} (${(0, bytes_1.default)(restSize)})`);
                    }
                    return files.map(format).join(', ');
                };
                rows.push({
                    name: '↪ ' + name,
                    files: getFiles(),
                    size: bytesAcc,
                    percent,
                });
            }
        }
        const maxLen = Math.max(...rows.map((r) => r.files.length));
        if (maxLen < 10) {
            // spread table
            leadingRow.files = '&nbsp; '.repeat(10);
        }
        const code = (text) => `\`${text}\``;
        for (const row of rows.sort((a, b) => b.size - a.size)) {
            const cells = [code(row.name), row.files, (0, bytes_1.default)(row.size), row.percent];
            lines.push(`| ${cells.join(' | ')} |`);
        }
        markdown.push(lines.join('\n'));
    }
    return markdown.join('\n\n');
}
exports.default = analyzeMetafile;
